import { Environment, Module, Node, Send } from '../model';
import { TypeSystemProblem, WollokParametricType, WollokType } from './wollokTypes';
export declare function newTypeVariables(env: Environment): Map<Node, TypeVariable>;
export declare function newSyntheticTVar(node?: Node): TypeVariable;
export declare function typeVariableFor(node: Node): TypeVariable;
export declare class TypeVariable {
    node: Node;
    typeInfo: TypeInfo;
    subtypes: TypeVariable[];
    supertypes: TypeVariable[];
    messages: Send[];
    cachedParams: Map<string, TypeVariable>;
    synthetic: boolean;
    hasProblems: boolean;
    constructor(node: Node);
    type(): WollokType;
    atParam(name: string): TypeVariable;
    newInstance(name: string): TypeVariable;
    instanceFor(instance: TypeVariable, send?: TypeVariable, name?: string): TypeVariable;
    hasMinType(type: WollokType): boolean;
    hasMaxType(type: WollokType): boolean;
    hasType(type: WollokType): boolean;
    setType(type: WollokType, closed?: boolean): this;
    addMinType(type: WollokType): void;
    addMaxType(type: WollokType): void;
    beSubtypeOf(tVar: TypeVariable): this;
    beSupertypeOf(tVar: TypeVariable): this;
    unify(tVar: TypeVariable): void;
    hasSubtype(tVar: TypeVariable): boolean;
    hasSupertype(tVar: TypeVariable): boolean;
    addSend(send: Send): void;
    hasSend(send: Send): boolean;
    allMinTypes(): WollokType[];
    allMaxTypes(): WollokType[];
    allPossibleTypes(): WollokType[];
    hasTypeInfered(): boolean;
    validSubtypes(): TypeVariable[];
    validSupertypes(): TypeVariable[];
    addSubtype(tVar: TypeVariable): void;
    addSupertype(tVar: TypeVariable): void;
    addProblem(problem: TypeSystemProblem): void;
    beSynthetic(): this;
    get closed(): boolean;
    toString(): string;
}
declare class TypeInfo {
    minTypes: WollokType[];
    maxTypes: WollokType[];
    closed: boolean;
    type(): WollokType;
    setType(type: WollokType, closed?: boolean): void;
    addMinType(type: WollokType): void;
    addMaxType(type: WollokType): void;
}
export declare function typeForModule(m: Module): WollokParametricType;
export {};
